# -*- coding: utf-8 -*-

from collective.behavior.talcondition.utils import _evaluateExpression
from collective.contact.plonegroup.utils import get_all_suffixes
from collective.contact.plonegroup.utils import get_organizations
from collective.contact.plonegroup.utils import get_plone_group_id
from eea.facetednavigation.browser.app.view import FacetedContainerView
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.content import uuidsToObjects
from imio.helpers.security import fplog
from imio.history.utils import getLastWFAction
from imio.pyutils.utils import sort_by_indexes
from plone import api
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from plone.dexterity.browser.view import DefaultView
from plone.supermodel.directives import FIELDSETS_KEY
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.PloneMeeting.config import ITEM_INSERT_METHODS
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.content.meeting import get_all_usable_held_positions
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.content.meeting import Meeting
from Products.PloneMeeting.MeetingConfig import POWEROBSERVERPREFIX
from Products.PloneMeeting.utils import _base_extra_expr_ctx
from Products.PloneMeeting.utils import field_is_empty
from Products.PloneMeeting.utils import get_attendee_short_title
from Products.PloneMeeting.utils import isPowerObserverForCfg
from Products.PloneMeeting.utils import redirect
from z3c.form.contentprovider import ContentProviders
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.interfaces import IFieldsAndContentProvidersForm
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.contentprovider.provider import ContentProviderBase
from zope.i18n import translate
from zope.interface import implements


def manage_fields(the_form):
    """
        Wipeout not enabled optional fields and fields
        for which condition is False
    """
    to_remove = []

    extra_expr_ctx = _base_extra_expr_ctx(the_form.context, {'view': the_form})
    for field_name, field_info in Meeting.FIELD_INFOS.items():
        if field_info['optional'] and \
           not the_form.show_field(field_name):
            to_remove.append(field_name)
        elif field_info['condition'] and \
                not _evaluateExpression(the_form.context,
                                        expression=field_info['condition'],
                                        extra_expr_ctx=extra_expr_ctx,
                                        raise_on_error=True,
                                        trusted=True):
            to_remove.append(field_name)

    # now remove fields
    for group in [the_form] + the_form.groups:
        for field_name in group.fields:
            if field_name in to_remove:
                group.fields = group.fields.omit(field_name)


def reorder_groups(the_form):
    """Reorder groups/fieldsets because when adding custom fields to an existing
       group, fieldset is duplicated and when merged for display,
       the overrided fieldset is displayed first."""
    order = {fieldset.__name__: i for i, fieldset in
             enumerate(IMeeting.getTaggedValue(FIELDSETS_KEY))}
    # in case group does not exist (new group added by custom plugin)
    # it will be displayed at the end
    indexes = [order.get(group.__name__, 9) for group in the_form.groups]
    # avoid useless reorder
    if indexes != sorted(indexes):
        the_form.groups = sort_by_indexes(the_form.groups, indexes)


def manage_label_assembly(the_form):
    """Depending on the fact that we use "assembly" alone or
       "assembly, excused, absents", we will translate the "assembly" label
       a different way.
    """
    widgets = getattr(the_form, 'w', None) or the_form.widgets
    if 'assembly' in widgets:
        if 'assembly_excused' in widgets or \
           'assembly_absents' in widgets:
            widgets['assembly'].label = _('title_attendees')


def manage_committees(the_form):
    """Depending on configuration, hide not used optional columns."""
    # not using committees?
    if "committees" not in the_form.used_attrs:
        return

    hidden_columns = []
    widget = the_form.w.get('committees')

    # check what columns to hide
    for optional_column in Meeting.FIELD_INFOS['committees']['optional_columns']:
        if not the_form.show_datagrid_column(widget, "committees", optional_column):
            hidden_columns.append(optional_column)

    # special behavior for assembly/attendees and signatures/signatories
    # in case config was switched if both fields are shown,
    # we keep the one not in the config
    if "assembly" not in hidden_columns and "attendees" not in hidden_columns:
        if "assembly" in the_form.used_attrs:
            hidden_columns.append("attendees")
        else:
            hidden_columns.append("assembly")
    if "signatures" not in hidden_columns and "signatories" not in hidden_columns:
        if "signatures" in the_form.used_attrs:
            hidden_columns.append("signatories")
        else:
            hidden_columns.append("signatures")

    # hide columns
    for column in widget.columns:
        if column['name'] in hidden_columns:
            column['mode'] = HIDDEN_MODE

    # hide widgets in rows
    for hidden_column_name in hidden_columns:
        for row in widget.widgets:
            for wdt in row.subform.widgets.values():
                if wdt.__name__ == hidden_column_name:
                    wdt.mode = HIDDEN_MODE


def manage_field_attendees(the_form):
    """Move ContentProvider from the_form.widgets to the 'assembly' group."""
    if the_form.show_attendees_fields() or \
       not the_form.show_field("assembly"):
        the_form.groups[1].widgets._data_keys.insert(0, "attendees")
        the_form.groups[1].widgets._data_values.insert(0, the_form.widgets["attendees_edit_provider"])
    # remove "attendees_edit_provider" from the_form.widgets
    # as it was either moved just here above or is not used
    the_form.widgets._data_keys.data.remove("attendees_edit_provider")
    the_form.widgets._data_values = [v for v in the_form.widgets._data_values
                                     if not v.__name__ == "attendees_edit_provider"]


def manage_groups(the_form):
    """Hide empty groups (fieldsets)."""
    groups = the_form.groups
    groups = [group for group in the_form.groups if group.widgets]
    the_form.groups = groups


class BaseMeetingView(object):
    """ """

    def shown_assembly_fields(self):
        '''Return the list of shown assembly field :
           - used assembly fields;
           - not empty assembly fields.'''
        cached_value = getattr(self, '_shown_assembly_fields', None)
        if cached_value is None:
            # get assembly fields
            field_names = self.cfg._assembly_field_names()
            return [field_name for field_name in field_names
                    if self.show_field(field_name)]

    def show_votes_observations(self):
        '''Show the votes_observations field to
           meeting managers and power observers.'''
        res = self.tool.isManager(self.cfg)
        if not res:
            res = isPowerObserverForCfg(self.cfg) or \
                (self.context.__class__.__name__ == 'Meeting' and
                 self.context.adapted().is_decided())
        return res

    def show_field(self, field_name):
        '''Show the p_field_name field?
           Must be enabled or not empty.'''
        is_used = field_name in self.used_attrs
        if not is_used and self.context.__class__.__name__ == 'Meeting':
            value = getattr(self.context, field_name, None)
            # special case for place that rely also on place_other
            if field_name == "place":
                if value != "other" or self.context.place_other is not None:
                    is_used = True
            elif value not in (None, -1, False):
                # None for empty RichTextFields
                # -1 for meeting_number/first_item_number
                # False for boolean fields
                is_used = True
        return is_used

    def show_datagrid_column(self, widget, field_name, column_name):
        '''Show the p_column_name or p_field_name DataGridField?
           Must be enabled or not empty.'''
        res = True
        used_attr_name = "{0}_{1}".format(field_name, column_name)
        if used_attr_name not in self.used_attrs and \
           field_is_empty(widget, column_name):
            res = False
        return res

    def show_attendees_fields(self):
        '''Display attendee related fields in view/edit?'''
        # caching
        show = getattr(self, "_show_attendees_fields_cache", None)
        if show is not None:
            return show
        # new meeting or existing meeting then manage history
        show = (self.context.__class__.__name__ != 'Meeting' and 'attendees' in self.used_attrs) or \
               (self.context.__class__.__name__ == 'Meeting' and
                (self.context.get_attendees() or not self.show_field("assembly")))
        show = bool(show)
        setattr(self, "_show_attendees_fields_cache", show)
        return show

    def _is_rich(self, widget):
        """Does given p_widget use a RichText field?"""
        return widget.field.__class__.__name__ == 'RichText' and \
            widget.__class__.__name__ != 'PMTextAreaWidget'

    def _is_datagrid(self, widget):
        """Does given p_widget use a DataGrid field?"""
        return widget.__class__.__name__ == 'BlockDataGridField'

    def view_widget(self, widget, empty_value="-"):
        """Render an empty_value instead nothing when field empty."""
        value = getattr(self.context, widget.__name__, None)
        if value is None and not self._is_rich(widget):
            rendered = "-"
        else:
            rendered = widget.render()
        return rendered

    def is_fieldset_not_empty(self, fieldset):
        """Is there a field of given p_fieldset that is not empty?"""
        is_not_empty = False
        for widget in self.fieldsets[fieldset].widgets.values():
            is_empty = field_is_empty(widget)
            if not is_empty:
                is_not_empty = True
                break
        return is_not_empty


class MeetingDefaultView(DefaultView, BaseMeetingView):
    """ """

    def updateFieldsFromSchemata(self):
        super(MeetingDefaultView, self).updateFieldsFromSchemata()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_attrs = self.cfg.getUsedMeetingAttributes()
        manage_fields(self)

    def _update(self):
        super(MeetingDefaultView, self)._update()
        manage_label_assembly(self)
        manage_committees(self)


def get_default_attendees(cfg):
    '''The default attendees are the active held_positions
       with 'present' in defaults.'''
    res = [held_pos.UID() for held_pos in get_all_usable_held_positions(cfg)
           if held_pos.defaults and 'present' in held_pos.defaults]
    return res


def get_default_signatories(cfg):
    '''The default signatories are the active held_positions
       with a defined signature_number.'''
    res = {}
    if "signatories" in cfg.getUsedMeetingAttributes():
        signers = [held_pos for held_pos in get_all_usable_held_positions(cfg)
                   if held_pos.defaults and 'present' in
                   held_pos.defaults and held_pos.signature_number]
        res = {signer.UID(): signer.signature_number for signer in signers}
    return res


def get_default_voters(cfg):
    '''The default voters are the active held_positions
       with 'voter' in defaults.'''
    res = []
    if cfg.getUseVotes():
        res = [held_pos.UID() for held_pos in get_all_usable_held_positions(cfg)
               if held_pos.defaults and 'voter' in held_pos.defaults]
    return res


class AttendeesEditProvider(ContentProviderBase, BaseMeetingView):

    template = \
        ViewPageTemplateFile('templates/meeting_attendees_edit.pt')

    def __init__(self, context, request, view):
        super(AttendeesEditProvider, self).__init__(context, request, view)
        self.__parent__ = view
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_attrs = self.cfg.getUsedMeetingAttributes()
        self.portal_url = api.portal.get().absolute_url()

    def render(self):
        return self.template()

    def get_all_users(self):
        """ """
        return get_all_usable_held_positions(self.context)

    def get_attendees(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        attendees = []
        if is_meeting:
            attendees = self.context.get_attendees()
        else:
            attendees = get_default_attendees(self.cfg)
        return attendees

    def get_excused(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        excused = []
        if is_meeting:
            excused = self.context.get_excused()
        return excused

    def get_absents(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        absents = []
        if is_meeting:
            absents = self.context.get_absents()
        return absents

    def get_signatories(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        signatories = []
        if is_meeting:
            signatories = self.context.get_signatories()
        else:
            signatories = get_default_signatories(self.cfg)
        return signatories

    def get_user_replacements(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        replacers = []
        if is_meeting:
            replacers = self.context.get_user_replacements()
        return replacers

    def get_voters(self):
        """ """
        is_meeting = self.context.__class__.__name__ == 'Meeting'
        voters = []
        if is_meeting:
            voters = self.context.get_voters()
        else:
            voters = get_default_voters(self.cfg)
        return voters

    def checked(self, cbid, muid, stored_value, req_key="meeting_attendees"):
        """Helper to manage checked checkboxes, especially when an error
           occured and the form is reloaded."""
        # case reloaded, check if checked in request
        if req_key in self.request:
            is_checked = cbid in self.request.get(req_key, [])
        # initial form load, nothing in request
        else:
            is_checked = muid in stored_value
        return is_checked

    def disabled(self, attendee_id, muid, stored_value, req_key="meeting_attendees"):
        """Helper to manage checked checkboxes, especially when an error
           occured and the form is reloaded."""
        # case reloaded, check if checked in request
        if req_key in self.request:
            is_disabled = attendee_id not in self.request.get(req_key, [])
        # initial form load, nothing in request
        else:
            is_disabled = muid not in stored_value
        return is_disabled

    def get_attendee_short_title(self, hp, cfg):
        """Helper to manage attendee short title."""
        return get_attendee_short_title(hp, cfg)


class MeetingEdit(DefaultEditForm, BaseMeetingView):
    """
        Edit form redefinition to customize fields.
    """
    implements(IFieldsAndContentProvidersForm)
    contentProviders = ContentProviders()
    contentProviders['attendees_edit_provider'] = AttendeesEditProvider
    # defining a contentProvider position is mandatory...
    contentProviders['attendees_edit_provider'].position = 0

    def updateFields(self):
        super(MeetingEdit, self).updateFields()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_attrs = self.cfg.getUsedMeetingAttributes()
        manage_fields(self)
        reorder_groups(self)

    def update(self):
        super(MeetingEdit, self).update()
        # shortcut 'widget' dictionary for all fieldsets
        # like in plone.autoform
        self.w = {}
        for group in self.groups:
            for k, v in group.widgets.items():
                self.w[k] = v
        manage_label_assembly(self)
        manage_field_attendees(self)
        manage_committees(self)
        manage_groups(self)


class MeetingAddForm(DefaultAddForm, BaseMeetingView):

    implements(IFieldsAndContentProvidersForm)
    contentProviders = ContentProviders()
    contentProviders['attendees_edit_provider'] = AttendeesEditProvider
    # defining a contentProvider position is mandatory...
    contentProviders['attendees_edit_provider'].position = 0

    def updateFields(self):
        super(MeetingAddForm, self).updateFields()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_attrs = self.cfg.getUsedMeetingAttributes()
        manage_fields(self)
        reorder_groups(self)

    def update(self):
        super(MeetingAddForm, self).update()
        # shortcut 'widget' dictionary for all fieldsets
        # like in plone.autoform
        self.w = {}
        for group in self.groups:
            for k, v in group.widgets.items():
                self.w[k] = v
        manage_label_assembly(self)
        manage_field_attendees(self)
        manage_committees(self)
        manage_groups(self)


class MeetingAdd(DefaultAddView):

    form = MeetingAddForm


class BaseMeetingFacetedView(FacetedContainerView):

    def __init__(self, context, request):
        """ """
        super(BaseMeetingFacetedView, self).__init__(context, request)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self._canonical = '<NOT SET>'


class MeetingFacetedAvailableItemsView(BaseMeetingFacetedView):
    """ """


class MeetingFacetedView(BaseMeetingFacetedView):
    """ """

    def _init(self):
        """ """
        # initialize member in call because it is Anonymous in __init__ of view...
        self.member = api.user.get_current()
        self.is_manager = self.tool.isManager(self.cfg)
        # make the 'view' widget available on faceted view
        view = self.context.restrictedTraverse('@@view')
        view.update()
        self.meeting_view = view

    def __call__(self):
        """ """
        self._init()
        return super(MeetingFacetedView, self).__call__()

    def show_page(self):
        '''If PloneMeeting is in "Restrict users" mode, the "Meeting view" page
           must not be shown to some users: users that do not have role
           MeetingManager and are not listed in a specific list.'''
        restrictMode = self.tool.getRestrictUsers()
        res = True
        if restrictMode:
            if not self.is_manager:
                # Check if the user is in specific list
                if self.member.getId() not in [
                        u.strip() for u in self.tool.getUnrestrictedUsers().split('\n')]:
                    res = False
        return res

    def _display_available_items_to(self):
        """Check if current user profile is selected in MeetingConfig.displayAvailableItemsTo."""
        displayAvailableItemsTo = self.cfg.getDisplayAvailableItemsTo()
        suffixes = []
        groups = []
        res = False
        cfgId = self.cfg.getId()
        for value in displayAvailableItemsTo:
            if value == 'app_users':
                suffixes = get_all_suffixes()
            elif value.startswith(POWEROBSERVERPREFIX):
                groups.append(get_plone_group_id(cfgId, value.split(POWEROBSERVERPREFIX)[1]))
        if suffixes:
            res = self.tool.userIsAmong(suffixes)
        if not res and groups:
            res = bool(set(groups).intersection(get_plone_groups_for_user()))
        return res

    def show_available_items(self):
        """Show the available items part?"""
        return (
            _checkPermission(ModifyPortalContent, self.context) or
            self._display_available_items_to()) and \
            self.context.wfConditions().may_accept_items()

    def warn_assembly(self, using_attendees=False):
        """Check if need to draw attention to the assembly fields to MeetingManagers.
           Warn if :
           - when using signatories, sufficient number of signatories must be defined;
           - when using assembly/signatures, assembly must be encoded and
             signatures must contain at least relevant lines of data."""
        warn = False
        if self.is_manager:
            if using_attendees:
                signature_numbers = self.context.get_signatories().values()
                # if we do not have signatures '1' and '2' we go further
                if u'1' not in signature_numbers or u'2' not in signature_numbers:
                    # double check, maybe we are in a case
                    # where we only have one signatory
                    default_signatories = get_default_signatories(self.cfg)
                    if sorted(default_signatories.values()) != sorted(signature_numbers):
                        warn = True
            else:
                if (len(self.context.get_signatures().split('\n')) <
                    len(self.cfg.getSignatures().split('\n'))) or \
                   not self.context.get_assembly(for_display=False, striked=False):
                    warn = True
        return warn


class MeetingInsertingMethodsHelpMsgView(BrowserView):
    """ """

    def __init__(self, context, request):
        """Initialize relevant data in index instead __init__
           because errors are hidden when occuring in __init__..."""
        self.context = context
        self.request = request
        self.portal_url = api.portal.get().absolute_url()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    def __call__(self):
        """Do business code in __call__ because errors are swallowed in __init__."""
        self.inserting_methods_fields_mapping = ITEM_INSERT_METHODS.copy()
        self.inserting_methods_fields_mapping.update(
            self.cfg.adapted().extraInsertingMethods().copy())
        return super(MeetingInsertingMethodsHelpMsgView, self).__call__()

    def fieldsToDisplay(self):
        """Depending on used inserting methods, display relevant fields."""
        res = []
        for method in self.cfg.getInsertingMethodsOnAddItem():
            for mapping in self.inserting_methods_fields_mapping[method['insertingMethod']]:
                if mapping.startswith('field_'):
                    res.append(mapping[6:])
        return res

    def orderedOrgs(self):
        """Display organizations if one of the selected inserting methods relies on organizations.
           Returns a list of tuples, with organization title as first element and
           goupsInCharge organizations titles as second element."""
        res = []
        orgs_inserting_methods = [
            method['insertingMethod'] for method in self.cfg.getInsertingMethodsOnAddItem()
            if 'organization' in self.inserting_methods_fields_mapping[method['insertingMethod']]]
        if orgs_inserting_methods:
            orgs = get_organizations(only_selected=True)
            res = [(org.Title(), ', '.join([gic.Title() for gic in org.get_groups_in_charge(the_objects=True)] or ''))
                   for org in orgs]
        return res

    def orderedCategories(self):
        """Display categories if one of the selected inserting methods relies on it."""
        categories = []
        categories_inserting_methods = [
            method['insertingMethod'] for method in self.cfg.getInsertingMethodsOnAddItem()
            if 'category' in self.inserting_methods_fields_mapping[method['insertingMethod']]]
        if categories_inserting_methods:
            categories = self.cfg.getCategories()
        return categories

    def orderedClassifiers(self):
        """Display classifiers if one of the selected inserting methods relies on it."""
        classifiers = []
        classifiers_inserting_methods = [
            method['insertingMethod'] for method in self.cfg.getInsertingMethodsOnAddItem()
            if 'classifier' in self.inserting_methods_fields_mapping[method['insertingMethod']]]
        if classifiers_inserting_methods:
            classifiers = self.cfg.getCategories(catType='classifiers')
        return classifiers


class MeetingUpdateItemReferences(BrowserView):
    """Call Meeting.update_item_references from a meeting."""

    def index(self):
        """ """
        self.context.update_item_references()
        msg = _('References of contained items have been updated.')
        api.portal.show_message(msg, request=self.request)
        return redirect(self.request, self.context.absolute_url())


class MeetingReorderItems(BrowserView):
    """Reorder items on the meeting depending on itemInsertOrder."""

    def _recompute_items_order(self):
        """Get every items and order it by get_item_insert_order."""
        items = self.context.get_items(ordered=True)
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        # sort items by insertOrder then by date it was presented
        # so items with same insert order will be sorted by WF transition 'present' time
        items = sorted([
            (self.context.get_item_insert_order(item, cfg), getLastWFAction(item, 'present')['time'], item)
            for item in items]
        )
        # get items
        items = [item[2] for item in items]
        # set items itemNumber
        itemNumber = 100
        self._orig_order = []
        for item in items:
            self._orig_order.append(item.getItemNumber(for_display=True))
            item.setItemNumber(itemNumber)
            itemNumber = itemNumber + 100
        self.context._finalize_item_insert(items_to_update=items)

    def index(self):
        """ """
        self._recompute_items_order()
        msg = _('Items have been reordered.')
        api.portal.show_message(msg, request=self.request)
        # add logging message to fingerpointing log
        extras = 'object={0} original_order={1}'.format(
            repr(self.context),
            ','.join(self._orig_order))
        fplog('meeting_items_reorder', extras=extras)
        return redirect(self.request, self.context.absolute_url())


class PresentSeveralItemsView(BrowserView):
    """
      This manage the view that presents several items into a meeting
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, uids):
        """ """
        wfTool = api.portal.get_tool('portal_workflow')
        # defer call to Meeting.update_item_references
        self.request.set('defer_Meeting_update_item_references', True)
        lowest_itemNumber = 0
        objs = uuidsToObjects(uids, ordered=True)
        for obj in objs:
            try:
                wfTool.doActionFor(obj, 'present')
            except WorkflowException:
                # sometimes an item may not be presented,
                # even if shown in presentable items
                api.portal.show_message(
                    _('Item \"${item_title}\" could not be presented!',
                      mapping={'item_title': safe_unicode(obj.Title())}),
                    request=self.request,
                    type='warning')
            if not lowest_itemNumber or obj.getItemNumber() < lowest_itemNumber:
                lowest_itemNumber = obj.getItemNumber()
        self.request.set('defer_Meeting_update_item_references', False)
        # now we may call update_item_references
        self.context.update_item_references(start_number=lowest_itemNumber)
        msg = translate('present_several_items_done',
                        domain='PloneMeeting',
                        context=self.request)
        plone_utils = api.portal.get_tool('plone_utils')
        plone_utils.addPortalMessage(msg)


class RemoveSeveralItemsView(BrowserView):
    """
      This manage the view that removes several items from a meeting
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, uids):
        """ """
        wfTool = api.portal.get_tool('portal_workflow')
        # make sure we have a list of uids, in some case, as it is called
        # by jQuery, we receive only one uid, as a string...
        if isinstance(uids, str):
            uids = [uids]
        # defer call to Meeting.update_item_references
        self.request.set('defer_Meeting_update_item_references', True)
        lowest_itemNumber = 0
        objs = uuidsToObjects(uids, ordered=True, unrestricted=True)
        for obj in objs:
            # save lowest_itemNumber for call to Meeting.update_item_references here under
            if not lowest_itemNumber or obj.getItemNumber() < lowest_itemNumber:
                lowest_itemNumber = obj.getItemNumber()
            # execute every 'back' transitions until item is in state 'validated'
            changedState = True
            while not obj.query_state() == 'validated':
                availableTransitions = [tr['id'] for tr in wfTool.getTransitionsFor(obj)]
                if not availableTransitions or not changedState:
                    break
                changedState = False
                # if several back transitions (like when WFAdaptation 'presented_item_back_to_xxx'
                # is selected), are available, give the priority to 'backToValidated'
                if 'backToValidated' in availableTransitions:
                    availableTransitions = ['backToValidated']
                for tr in availableTransitions:
                    if tr.startswith('back'):
                        wfTool.doActionFor(obj, tr)
                        changedState = True
                        break

        self.request.set('defer_Meeting_update_item_references', False)
        # now we may call update_item_references
        self.context.update_item_references(start_number=lowest_itemNumber)
        msg = translate('remove_several_items_done',
                        domain='PloneMeeting',
                        context=self.request)
        plone_utils = api.portal.get_tool('plone_utils')
        plone_utils.addPortalMessage(msg)



# -*- coding: utf-8 -*-

from imio.helpers.setup import load_type_from_package
from imio.webspellchecker.config import set_disable_autosearch_in
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator


class Migrate_To_4215(Migrator):

    def _updateConfigCustomAdvisersDataGrid(self):
        """MeetingConfig.customAdvisers get a new column "is_delay_calendar_days"."""
        logger.info('Updating datagridfield "customAdvisers" for every MeetingConfigs...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            custom_advisers = cfg.getCustomAdvisers()
            for ca in custom_advisers:
                if "is_delay_calendar_days" not in ca:
                    ca["is_delay_calendar_days"] = "0"
            cfg.setCustomAdvisers(custom_advisers)
        logger.info('Done.')

    def _reloadMeetingConfigsForItemWorkflows(self):
        """Reload MeetingConfigs if using "return_to_proposing_group" with validation."""
        logger.info("Updating item WF using 'return_to_proposing_group with validation'...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if len([state_id for state_id in cfg.getItemWorkflow(True)
                    if state_id.startswith('returned_to_proposing_group')]) > 1:
                cfg.registerPortalTypes()
        logger.info('Done.')

    def _fixWSCConfigAndCleanBrokenAnnexes(self):
        """WSC could sometimes break quickupload leading to broken annexes.
           Adapt WSC config so it is no more used in quickupload and clean broken annexes."""
        logger.info("Updating WSC config and removing broken annexes...")
        if self.portal.portal_quickinstaller.isProductInstalled('imio.webspellchecker'):
            # disable WSC in quickupload
            set_disable_autosearch_in(
                u'["#form-widgets-title", "#form-widgets-description", '
                u'".select2-focusser", ".select2-input"]')
            # remove broken annexes
            self._removeBrokenAnnexes()
        logger.info('Done.')

    def _updateWFWriteMarginalNotesPermission(self):
        """Re-apply WFs (portal_setup and MeetingConfig) as WriteMarginalNotes
           permission is now available when item is "presented"."""
        logger.info("Updating workflows 'WriteMarginalNotes' permission...")
        # update workflow definition
        self.runProfileSteps('Products.PloneMeeting', steps=['workflow'], profile='default')
        # reload every MeetingConfigs so WF are updated
        self.reloadMeetingConfigs()
        # update every "presented" items
        for cfg in self.tool.objectValues('MeetingConfig'):
            item_wf = cfg.getItemWorkflow(True)
            for brain in self.catalog(portal_type=cfg.getItemTypeName(), review_state="presented"):
                item = brain.getObject()
                item_wf.updateRoleMappingsFor(item)
        logger.info('Done.')

    def run(self, extra_omitted=[], from_migration_to_4200=False):

        logger.info('Migrating to PloneMeeting 4215...')
        if not from_migration_to_4200:
            # this will upgrade collective.contact.core especially
            # that reinstalls imio.fpaudit, that will itself reinstall collective.documentgenerator
            self.upgradeAll(omit=['Products.PloneMeeting:default',
                                  self.profile_name.replace('profile-', '')])
            # reload ConfigurablePODTemplate
            load_type_from_package('ConfigurablePODTemplate', 'Products.PloneMeeting:default')
            # hide document-generation-link default viewlet
            self.ps.runImportStepFromProfile('profile-Products.PloneMeeting:default', 'viewlets')
        self._updateConfigCustomAdvisersDataGrid()
        self._reloadMeetingConfigsForItemWorkflows()
        self._fixWSCConfigAndCleanBrokenAnnexes()
        self._updateWFWriteMarginalNotesPermission()
        logger.info('Migrating to PloneMeeting 4215... Done.')


def migrate(context):
    '''This migration function will:

       1) Upgrade all and make sure documentgenerator overrides are re-applied;
       2) Update MeetingConfig.customAdvisers to add new column "is_delay_calendar_days";
       3) Reload MeetingConfigs if using "return_to_proposing_group" with validation;
       4) Fix WSC config and remove broken annexes;
       5) Update "WriteMarginalNotes" for every item WF and "presented" items.
    '''
    migrator = Migrate_To_4215(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from imio.helpers.content import get_user_fullname
from imio.history.interfaces import IImioHistory
from imio.history.utils import add_event_to_history
from plone import api
from plone.z3cform.layout import wrap_form
from Products.Archetypes import DisplayList
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.PloneMeeting.config import PMMessageFactory as _
from z3c.form import button
from z3c.form import field
from z3c.form import form
from zope import interface
from zope import schema
from zope.component import getAdapter
from zope.component.hooks import getSite
from zope.i18n import translate


class ItemEmergencyView(BrowserView):
    '''Render the item emergency HTML on the meetingitem_view.'''

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request
        self.portal_url = getToolByName(self, 'portal_url').getPortalObject().absolute_url()

    def listSelectableEmergencies(self):
        '''Returns a list of emergencies the current user can set the item to.'''
        # we base our actions on avaialble terms in the MeetingItem.emergency field vocabulary
        emergencies = self.context.listEmergencies()
        emergencyKeys = emergencies.keys()
        currentEmergency = self.context.getEmergency()
        # now check if user can ask emergency if it is not already the case
        if not self.context.adapted().mayAskEmergency():
            emergencyKeys.remove('emergency_asked')
            emergencyKeys.remove('no_emergency')
        # now check if user can accept/refuse and asked emergency if it is not already the case
        if not self.context.adapted().mayAcceptOrRefuseEmergency() or currentEmergency == 'no_emergency':
            emergencyKeys.remove('emergency_accepted')
            emergencyKeys.remove('emergency_refused')
        elif not currentEmergency == 'emergency_asked':
            if currentEmergency in emergencyKeys:
                emergencyKeys.remove(currentEmergency)
        # now if currentEmergency is still in emergencies, we remove it
        if currentEmergency in emergencyKeys:
            emergencyKeys.remove(currentEmergency)
        # now build a vocabulary with left values
        res = []
        for emergency in emergencies.items():
            if emergency[0] in emergencyKeys:
                res.append(emergency)
        return DisplayList(tuple(res))


class ItemEmergencyHistoryView(BrowserView):
    '''Display history of emergency value changes.'''
    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request
        self.tool = api.portal.get_tool('portal_plonemeeting')

    def getHistory(self, checkMayViewEvent=True, checkMayViewComment=True):
        """ """
        adapter = getAdapter(self.context, IImioHistory, 'emergency_changes')
        history = adapter.getHistory(
            checkMayViewEvent=checkMayViewEvent,
            checkMayViewComment=checkMayViewComment)
        if not history:
            return []
        history.sort(key=lambda x: x["time"], reverse=True)
        return history

    def renderComments(self, comments, mimetype='text/plain'):
        """
          Borrowed from imio.history.
        """
        transformsTool = api.portal.get_tool('portal_transforms')
        data = transformsTool.convertTo('text/x-html-safe', comments, mimetype=mimetype)
        return data.getData()

    def get_user_fullname(self, user_id):
        return get_user_fullname(user_id)


def new_emergency_value_default():
    """
      Get the value from the REQUEST as it is passed when calling the
      form : form?new_emergency_value=new_value.
    """
    request = getSite().REQUEST
    return request.get('new_emergency_value', '')


class IItemEmergencyComment(interface.Interface):
    comment = schema.Text(
        title=_(u"Comment"),
        description=_(u""),
        required=True,)

    new_emergency_value = schema.TextLine(
        title=_(u"New emergency value"),
        description=_(u""),
        defaultFactory=new_emergency_value_default)


class ItemEmergencyChangeForm(form.Form):
    """
      This form will give the possibility to add a
      required comment while changing item emergency.
    """
    label = _(u"Manage item emergency")
    description = u''

    fields = field.Fields(IItemEmergencyComment)
    ignoreContext = True  # don't use context to get widget data

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.label = translate(self.label,
                               domain='PloneMeeting',
                               context=self.request)

    @button.buttonAndHandler(_('save'), name='save_item_emergency')
    def handleSaveItemEmergency(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        # check that given 'new_emergency_value' is available in the field vocabulary
        # if not available, just raise Unauthorized
        new_emergency_value = data.get('new_emergency_value')
        itemEmergencyView = self.context.unrestrictedTraverse('@@item-emergency')
        if new_emergency_value not in itemEmergencyView.listSelectableEmergencies():
            raise Unauthorized
        self.context.setEmergency(new_emergency_value)
        # add a line to the item's emergency_change_history
        add_event_to_history(
            self.context,
            'emergency_changes_history',
            action=new_emergency_value,
            comments=data['comment'])
        # update item
        self.context._update_after_edit(idxs=[])
        plone_utils = api.portal.get_tool('plone_utils')
        plone_utils.addPortalMessage(_("Item emergency changed."))
        self.request.RESPONSE.redirect(self.context.absolute_url())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self.request.RESPONSE.redirect(self.context.absolute_url())

    def update(self):
        """ """
        # raise Unauthorized if current user can not manage itemAssembly
        if not self.context.mayQuickEdit('emergency'):
            raise Unauthorized
        super(ItemEmergencyChangeForm, self).update()
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')

    def updateWidgets(self):
        # hide field 'new_emergency_value'
        self.fields['new_emergency_value'].mode = 'hidden'
        translate_new_emergency_value = translate(new_emergency_value_default(),
                                                  domain='PloneMeeting',
                                                  context=self.request)
        self.fields['comment'].field.description = translate(
            'change_emergency_descr',
            domain='PloneMeeting',
            mapping={'new_emergency_value': translate_new_emergency_value},
            context=self.request,
            default=u"You are about to change emergency for this item to <span style='font-weight: bold;'>"
            u"${new_emergency_value}</span>, please enter a comment.")
        super(ItemEmergencyChangeForm, self).updateWidgets()


ItemEmergencyChangeFormWrapper = wrap_form(ItemEmergencyChangeForm)



# -*- coding: utf-8 -*-

from Acquisition import aq_base
from cPickle import dumps
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import get_plone_groups_for_user
from plone import api
from plone.app.querystring import queryparser
from plone.memoize import ram
from Products.Archetypes.BaseObject import BaseObject
from Products.Archetypes.Field import Field
from Products.Archetypes.Field import TextField
from Products.CMFCore.TypesTool import TypeInformation
from Products.PloneMeeting import logger
from Products.PlonePAS.tools.membership import MembershipTool
from Products.PortalTransforms.cache import Cache
from Products.PortalTransforms.transforms import safe_html
from Products.PortalTransforms.transforms.safe_html import CSS_COMMENT
from Products.PortalTransforms.transforms.safe_html import decode_htmlentities
from time import time
from types import StringType
from z3c.form import interfaces
from z3c.form.widget import SequenceWidget
from zope.i18nmessageid import Message
from zope.ramcache.ram import Storage


def _patched_equal(context, row):
    """Monkeypatch _equal so we never pass None as values or it breaks as
       None can not be indexed.
    """
    row_values = row.values
    if row_values is None:
        row_values = []
    return {row.index: {'query': row_values, }}


queryparser.__pm_old_equal = queryparser._equal
queryparser._equal = _patched_equal
logger.info("Monkey patching plone.app.querystring.queryparser (_equal)")
_equal = _patched_equal


def userAndGroupsAwarePortalTransformsCacheKey():
    """Monkeypatch Products.PortalTransforms.Cache._genCacheKey
       to generate a cache key that is aware of current user and it's groups.
    """

    def _genCacheKey(self, identifier, *args):
        # XXX begin changes by PM, do the cache key user and groups aware
        from plone import api
        from plone.api.exc import InvalidParameterError
        try:
            # while creating a new Plone Site, portal_plonemeeting is not available
            api.portal.get_tool('portal_plonemeeting')
        except InvalidParameterError:
            return Cache.__pm_old_genCacheKey(self, identifier, *args)
        # XXX end of changes by PM
        key = identifier
        for arg in args:
            key = '%s_%s' % (key, arg)
        key = key.replace('/', '_')
        key = key.replace('+', '_')
        key = key.replace('-', '_')
        key = key.replace(' ', '_')
        # XXX begin changes by PM, do the cache key user and groups aware
        user_id = get_current_user_id()
        groups = get_plone_groups_for_user()
        key = '%s_%s_%s' % (key, user_id, '_'.join(groups))
        # XXX end of changes by PM
        if hasattr(aq_base(self.context), 'absolute_url'):
            return key, self.context.absolute_url()
        return key

    Cache.__pm_old_genCacheKey = Cache._genCacheKey
    Cache._genCacheKey = _genCacheKey
    logger.info("Monkey patching Products.PortalTransforms.cache (_genCacheKey)")


userAndGroupsAwarePortalTransformsCacheKey()


BaseObject.__old_pm_validate = BaseObject.validate


def validate(self, REQUEST=None, errors=None, data=None, metadata=None):
    """Monkeypatch to log errors because sometimes, when errors occur in multiple
       or on disabled fields, it is not visible into the UI."""
    errors = self.__old_pm_validate(REQUEST, errors, data, metadata)
    if errors and api.user.get_current().has_role('Manager'):
        logger.info(errors)
    return errors


BaseObject.validate = validate
logger.info("Monkey patching Products.Archetypes.BaseObject.BaseObject (validate)")


def hasScript(s):
    """Override to keep data:image elements, turned 'data:' to 'data:text'
    """
    s = decode_htmlentities(s)
    s = s.replace('\x00', '')
    s = CSS_COMMENT.sub('', s)
    s = ''.join(s.split()).lower()
    for t in ('script:', 'expression:', 'expression(', 'data:text'):
        if t in s:
            return True
    return False


safe_html.hasScript = hasScript
logger.info("Monkey patching Products.PortalTransforms.transforms.safe_html (hasScript)")


Field.__old_pm_validate_content_types = Field.validate_content_types


def validate_content_types(self, instance, value, errors):
    """Avoid wrong validation error when html value is detected as text/plain,
       this may occur when having a image data:base64 (when using imagerotate)."""
    error = Field.__old_pm_validate_content_types(self, instance, value, errors)
    if error and "text/plain" in error:
        if isinstance(value, StringType) and \
           value.startswith('<p>'):
            errors.pop(self.getName())
            error = None
    return error


Field.validate_content_types = validate_content_types
logger.info("Monkey patching Products.Archetypes.Field.Field (validate_content_types)")


TextField.__old_pm__process_input = TextField._process_input


def _process_input(self, value, file=None, default=None,
                   mimetype=None, instance=None, **kwargs):
    """Initialize new field correctly with text/html content_type."""
    file, mimetype, filename = self.__old_pm__process_input(
        value=value, file=file, default=default, mimetype=mimetype, instance=instance, **kwargs)
    # this case if when a new field is initialized on an existing element
    if not instance.checkCreationFlag() and kwargs.get('_initializing_') and 'schema' in kwargs:
        field_name = kwargs['field'] if isinstance(kwargs['field'], str) else \
            kwargs['field'].__name__
        logger.info("Initializing new field \"{0}\" on existing element at {1}".format(
            field_name, '/'.join(instance.getPhysicalPath())))
        default_content_type = kwargs['schema'][field_name].default_content_type
        if default_content_type:
            file.setContentType(instance, default_content_type)
    return file, mimetype, filename


TextField._process_input = _process_input
logger.info("Monkey patching Products.Archetypes.Field.TextField (_process_input)")


def extract(self, default=interfaces.NO_VALUE):
    """See z3c.form.interfaces.IWidget."""

    if (self.name not in self.request and
            self.name + '-empty-marker' in self.request):
        return []
    value = self.request.get(self.name, default)
    if value != default:
        if not isinstance(value, (tuple, list)):
            value = (value,)
        # do some kind of validation, at least only use existing values
        for token in value:
            # XXX begin do not encode to utf-8 for MasterSelectWidget
            # as we use unicode values or validation fails with "field required"
            if isinstance(token, unicode) and not self.__class__.__name__ == 'MasterSelectWidget':
                token = token.encode('utf-8')
            # XXX end
            if token == self.noValueToken:
                continue
            try:
                self.terms.getTermByToken(token)
            except LookupError:
                return default
    return value


SequenceWidget.extract = extract
logger.info("Monkey patching z3c.form.widget.SequenceWidget (extract)")


def getMemberInfo_cachekey(method, self, memberId=None):
    '''cachekey method for self.getMemberInfo.
       Cache is invalidated by plone.app.controlpanel upon any control panel changes.'''
    return memberId


MembershipTool.__old_pm_getMemberInfo = MembershipTool.getMemberInfo


@ram.cache(getMemberInfo_cachekey)
def getMemberInfo(self, memberId=None):
    """Monkeypatched to add caching."""
    return self.__old_pm_getMemberInfo(memberId)


MembershipTool.getMemberInfo = getMemberInfo
logger.info("Monkey patching Products.PlonePAS.tools.membership.MembershipTool (getMemberInfo)")


Storage.__old_pm_getEntry = Storage.getEntry


def getEntry(self, ob, key):
    if self.lastCleanup <= time() - self.cleanupInterval:
        self.cleanup()

    try:
        data = self._data[ob][key]
    except KeyError:
        if ob not in self._misses:
            self._misses[ob] = 0
        self._misses[ob] += 1
        raise
    else:
        data[2] += 1                    # increment access count
        # XXX begin change by PM, update timestamp
        timestamp = time()
        data[1] = timestamp
        # XXX end change by PM

        return data[0]


Storage.getEntry = getEntry
logger.info("Monkey patching zope.ramcache.ram.Storage (getEntry)")


Storage.__old_pm_getStatistics = Storage.getStatistics


def getStatistics(self):
    objects = self._data.keys()
    objects.sort()
    result = []

    for ob in objects:
        size = len(dumps(self._data[ob]))
        hits = sum(entry[2] for entry in self._data[ob].itervalues())
        from DateTime import DateTime
        older_date = min(entry[1] for entry in self._data[ob].itervalues())
        result.append({'path': ob,
                       'hits': hits,
                       'misses': self._misses.get(ob, 0),
                       'size': size,
                       'entries': len(self._data[ob]),
                       'older_date': older_date and DateTime(older_date)})
    return tuple(result)


Storage.getStatistics = getStatistics
logger.info("Monkey patching zope.ramcache.ram.Storage (getStatistics)")


def Title(self):
    """Same code as dexterity's fti (DexterityFTI) for AT fti."""
    if self.title and self.i18n_domain:
        try:
            return Message(self.title.decode('utf8'), self.i18n_domain)
        except UnicodeDecodeError:
            return Message(self.title.decode('latin-1'), self.i18n_domain)
    else:
        return self.title or self.getId()


TypeInformation.__old_Title = TypeInformation.Title
TypeInformation.Title = Title



# -*- coding: utf-8 -*-
#
# File: MeetingGroup.py
#
# Copyright (c) 2015 by Imio.be
# Generator: ArchGenXML Version 2.7
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#

from AccessControl import ClassSecurityInfo
from imio.helpers.cache import invalidate_cachekey_volatile_for
from plone import api
from Products.Archetypes.atapi import AttributeStorage
from Products.Archetypes.atapi import BaseContent
from Products.Archetypes.atapi import BaseSchema
from Products.Archetypes.atapi import DisplayList
from Products.Archetypes.atapi import LinesField
from Products.Archetypes.atapi import MultiSelectionWidget
from Products.Archetypes.atapi import registerType
from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import SelectionWidget
from Products.Archetypes.atapi import StringField
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.DataGridField import DataGridField
from Products.DataGridField.Column import Column
from Products.DataGridField.SelectColumn import SelectColumn
from Products.PloneMeeting.config import PROJECTNAME
from Products.PloneMeeting.config import WriteRiskyConfig
from Products.PloneMeeting.utils import computeCertifiedSignatures
from Products.PloneMeeting.utils import createOrUpdatePloneGroup
from Products.PloneMeeting.utils import getCustomAdapter
from Products.PloneMeeting.utils import listifySignatures
from zope.i18n import translate
from zope.interface import implements

import interfaces


__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'

schema = Schema((

    StringField(
        name='acronym',
        widget=StringField._properties['widget'](
            label='Acronym',
            label_msgid='PloneMeeting_label_acronym',
            i18n_domain='PloneMeeting',
        ),
        required=True,
        write_permission="PloneMeeting: Write risky config",
    ),
    LinesField(
        name='itemAdviceStates',
        widget=MultiSelectionWidget(
            description="ItemAdviceStates",
            description_msgid="group_item_advice_states_descr",
            format="checkbox",
            label='Itemadvicestates',
            label_msgid='PloneMeeting_label_itemAdviceStates',
            i18n_domain='PloneMeeting',
        ),
        multiValued=1,
        vocabulary='listItemStates',
        write_permission="PloneMeeting: Write risky config",
    ),
    LinesField(
        name='itemAdviceEditStates',
        widget=MultiSelectionWidget(
            description="ItemAdviceEditStates",
            description_msgid="group_item_advice_edit_states_descr",
            format="checkbox",
            label='Itemadviceeditstates',
            label_msgid='PloneMeeting_label_itemAdviceEditStates',
            i18n_domain='PloneMeeting',
        ),
        multiValued=1,
        vocabulary='listItemStates',
        write_permission="PloneMeeting: Write risky config",
    ),
    LinesField(
        name='itemAdviceViewStates',
        widget=MultiSelectionWidget(
            description="ItemAdviceViewStates",
            description_msgid="group_item_advice_view_states_descr",
            format="checkbox",
            label='Itemadviceviewstates',
            label_msgid='PloneMeeting_label_itemAdviceViewStates',
            i18n_domain='PloneMeeting',
        ),
        multiValued=1,
        vocabulary='listItemStates',
        write_permission="PloneMeeting: Write risky config",
    ),
    StringField(
        name='keepAccessToItemWhenAdviceIsGiven',
        widget=SelectionWidget(
            format="select",
            description="KeepAccessToItemWhenAdviceIsGiven",
            description_msgid="group_keep_access_to_item_when_advice_is_given_descr",
            label='Keepaccesstoitemwhenadviceisgiven',
            label_msgid='PloneMeeting_label_keepAccessToItemWhenAdviceIsGiven',
            i18n_domain='PloneMeeting',
        ),
        enforceVocabulary=True,
        vocabulary='listKeepAccessToItemWhenAdviceIsGiven',
        write_permission="PloneMeeting: Write risky config",
    ),
    StringField(
        name='asCopyGroupOn',
        widget=StringField._properties['widget'](
            size=100,
            description="AsCopyGroupOn",
            description_msgid="as_copy_group_on_descr",
            label='Ascopygroupon',
            label_msgid='PloneMeeting_label_asCopyGroupOn',
            i18n_domain='PloneMeeting',
        ),
        write_permission="PloneMeeting: Write risky config",
    ),
    DataGridField(
        name='certifiedSignatures',
        widget=DataGridField._properties['widget'](
            columns={'signatureNumber':
                     SelectColumn("Certified signatures signature number",
                                  vocabulary="listSignatureNumbers",
                                  col_description="Select the signature number, keep signatures ordered by number."),
                     'name':
                     Column("Certified signatures signatory name",
                            col_description="Name of the signatory (for example 'Mister John Doe')."),
                     'function':
                     Column("Certified signatures signatory function",
                            col_description="Function of the signatory (for example 'Mayor')."),
                     'date_from':
                     Column("Certified signatures valid from (included)",
                            col_description="Enter valid from date, use following format : YYYY/MM/DD, "
                                            "leave empty so it is always valid."),
                     'date_to':
                     Column("Certified signatures valid to (included)",
                            col_description="Enter valid to date, use following format : YYYY/MM/DD, "
                                            "leave empty so it is always valid."), },
            label_msgid="PloneMeeting_label_group_certifiedSignatures",
            description="GroupCertifiedSignatures",
            description_msgid="group_certified_signatures_descr",
            label='Certifiedsignatures',
            i18n_domain='PloneMeeting',
        ),
        validators=('isValidCertifiedSignatures',),
        allow_oddeven=True,
        write_permission="PloneMeeting: Write harmless config",
        columns=('signatureNumber', 'name', 'function', 'date_from', 'date_to'),
        allow_empty_rows=False,
    ),
    LinesField(
        name='groupsInCharge',
        widget=MultiSelectionWidget(
            description="GroupsInCharge",
            description_msgid="groups_in_charge_descr",
            format="checkbox",
            label='Groupsincharge',
            label_msgid='PloneMeeting_label_groupsInCharge',
            i18n_domain='PloneMeeting',
        ),
        multiValued=1,
        vocabulary='listActiveMeetingGroups',
        write_permission="PloneMeeting: Write risky config",
    ),

),
)

MeetingGroup_schema = BaseSchema.copy() + \
    schema.copy()

MeetingGroup_schema['id'].write_permission = "PloneMeeting: Write risky config"
MeetingGroup_schema['title'].write_permission = "PloneMeeting: Write risky config"
MeetingGroup_schema.changeSchemataForField('description', 'default')
MeetingGroup_schema.moveField('description', after='title')
MeetingGroup_schema['description'].storage = AttributeStorage()
MeetingGroup_schema['description'].write_permission = "PloneMeeting: Write risky config"
MeetingGroup_schema['description'].widget.description = " "
MeetingGroup_schema['description'].widget.description_msgid = "empty_description"
# hide metadata fields and even protect it by the WriteRiskyConfig permission
for field in MeetingGroup_schema.getSchemataFields('metadata'):
    field.widget.visible = {'edit': 'invisible', 'view': 'invisible'}
    field.write_permission = WriteRiskyConfig


class MeetingGroup(BaseContent, BrowserDefaultMixin):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IMeetingGroup)

    meta_type = 'MeetingGroup'
    _at_rename_after_creation = True

    schema = MeetingGroup_schema

    security.declarePublic('query_state')

    def query_state(self):
        '''In what state am I ?'''
        return self.portal_workflow.getInfoFor(self, 'review_state')

    security.declarePrivate('listSignatureNumbers')

    def listSignatureNumbers(self):
        '''Vocabulary for column 'signatureNumber' of MeetingGroup.certifiedSignatures.'''
        res = []
        for number in range(1, 11):
            res.append((str(number), str(number)))
        return DisplayList(tuple(res))

    security.declarePrivate('listItemStates')

    def listItemStates(self):
        '''Lists the states of the item workflow for each MeetingConfig.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        res = []
        for cfg in tool.getActiveConfigs():
            cfgItemStates = cfg.listStates('Item')
            cfgId = cfg.getId()
            # cfgItemStates is a list of tuple, ready to move to a DisplayList
            for key, value in cfgItemStates:
                # build a strong id
                res.append(("%s__state__%s" % (cfgId, key),
                            "%s - %s" % (unicode(cfg.Title(include_config_group=True), 'utf-8'), value)))
        return DisplayList(tuple(res)).sortedByValue()

    security.declarePrivate('listKeepAccessToItemWhenAdviceIsGiven')

    def listKeepAccessToItemWhenAdviceIsGiven(self):
        '''Vocabulary for field keepAccessToItemWhenAdviceIsGiven.'''
        res = [('',
                translate('use_meetingconfig_value',
                          domain='PloneMeeting',
                          context=self.REQUEST)),
               ('0',
                translate('boolean_value_false',
                          domain='PloneMeeting',
                          context=self.REQUEST)),
               ('1',
                translate('boolean_value_true',
                          domain='PloneMeeting',
                          context=self.REQUEST)),
               ]
        return DisplayList(tuple(res))

    security.declarePrivate('listActiveMeetingGroups')

    def listActiveMeetingGroups(self):
        """
          Vocabulary for the groupsInCharge field.
          It returns every active MeetingGroups.
        """
        res = []
        tool = api.portal.get_tool('portal_plonemeeting')
        for mGroup in tool.getMeetingGroups():
            res.append((mGroup.getId(), mGroup.Title()))
        # make sure that if a configuration was defined for a group
        # that is now inactive, it is still displayed
        storedGroupsInCharge = self.getGroupsInCharge()
        if storedGroupsInCharge:
            groupsInVocab = [group[0] for group in res]
            for storedGroupInCharge in storedGroupsInCharge:
                if storedGroupInCharge not in groupsInVocab:
                    mGroup = getattr(tool, storedGroupInCharge)
                    res.append((mGroup.getId(), mGroup.Title()))
        return DisplayList(res).sortedByValue()

    def getPloneGroupId(self, suffix):
        '''Returns the id of the Plone group that corresponds to me and
           p_suffix.'''
        return '%s_%s' % (self.id, suffix)

    def getOrder(self, associatedGroupIds=None, onlyActive=True):
        '''At what position am I among all the active groups ? If
           p_associatedGroupIds is not None or empty, this method must return
           the order of the lowest group among all associated groups (me +
           associated groups).
           If p_onlyActive is True, only consider active groups, if not
           take also deactivated groups.'''
        groups = self.getParentNode().getMeetingGroups(onlyActive=onlyActive)
        i = groups.index(self)
        # if we received associatedGroupIds we must consider associated group
        # that has the lowest position
        if associatedGroupIds:
            # groups are sorted so, the first we find, we return it
            groupIds = [group.getId() for group in groups]
            for groupId in groupIds:
                if groupId in associatedGroupIds:
                    # we found the associatedGroup with lowest position, now check
                    # that the lowest position of this associated group is lower or not
                    # than the position of the proposing group
                    associatedGroupIndex = groupIds.index(groupId)
                    if associatedGroupIndex < i:
                        i = associatedGroupIndex
                    break
        return i

    security.declarePrivate('at_post_create_script')

    def at_post_create_script(self):
        ''' '''
        pass

    security.declarePrivate('at_post_edit_script')

    def at_post_edit_script(self):
        """ """
        pass

    def _invalidateCachedVocabularies(self):
        """Clean cache for vocabularies using MeetingGroups."""
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.proposinggroupsvocabulary")
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.proposinggroupacronymsvocabulary")
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.proposinggroupsforfacetedfiltervocabulary")
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.groupsinchargevocabulary")
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.askedadvicesvocabulary")

    def _createOrUpdatePloneGroup(self, groupSuffix):
        '''This will create the PloneGroup that corresponds to me
           and p_groupSuffix, if group already exists, it will just update it's title.'''
        groupId = self.getPloneGroupId(groupSuffix)
        groupTitle = self.Title()
        wasCreated = createOrUpdatePloneGroup(groupId=groupId, groupTitle=groupTitle, groupSuffix=groupSuffix)
        if wasCreated:
            portal_groups = api.portal.get_tool('portal_groups')
            portal_groups.setRolesForGroup(groupId, ('MeetingObserverGlobal',))

    security.declarePublic('getSelf')

    def getSelf(self):
        if self.getTagName() != 'MeetingGroup':
            return self.context
        return self

    security.declarePublic('adapted')

    def adapted(self):
        return getCustomAdapter(self)

    security.declareProtected('Modify portal content', 'onEdit')

    def onEdit(self, isCreated):
        '''See doc in interfaces.py.'''
        pass

    security.declarePublic('getItemAdviceStates')

    def getItemAdviceStates(self, cfg=None, **kwargs):
        '''This is an overridden version of the Archetypes accessor for field
           "itemAdviceStates". When called by Archetypes (with no arg), it
           simply returns the content of field MeetingGroup.itemAdviceStates.
           When called with a p_cfg (MeetingConfig), if MeetingGroup.itemAdviceStates
           is not empty it returns it, but manipulates returned value as stored value is
           something like 'meeting-config-sample__state__itemcreated' and we want 'itemcreated';
           else, it returns the global, default list in cfg.itemAdviceStates that correctly contains
           state values.'''
        res = self.getField('itemAdviceStates').get(self, **kwargs)

        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this MeetingGroup, use value defined on the cfg
            res = tmpres or cfg.getItemAdviceStates()
        return tuple(res)

    security.declarePublic('getItemAdviceEditStates')

    def getItemAdviceEditStates(self, cfg=None, **kwargs):
        '''See docstring of method MeetingGroup.getItemAdviceStates.'''
        res = self.getField('itemAdviceEditStates').get(self, **kwargs)
        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this MeetingGroup, use value defined on the cfg
            res = tmpres or cfg.getItemAdviceEditStates()
        return tuple(res)

    security.declarePublic('getItemAdviceViewStates')

    def getItemAdviceViewStates(self, cfg=None, **kwargs):
        '''See docstring of method MeetingGroup.getItemAdviceStates.'''
        res = self.getField('itemAdviceViewStates').get(self, **kwargs)
        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this MeetingGroup, use value defined on the cfg
            res = tmpres or cfg.getItemAdviceViewStates()
        return tuple(res)

    security.declarePublic('getKeepAccessToItemWhenAdviceIsGiven')

    def getKeepAccessToItemWhenAdviceIsGiven(self, cfg=None, **kwargs):
        '''This is an overridden version of the Archetypes accessor for field
           "keepAccessToItemWhenAdviceIsGiven". When called by Archetypes (with no arg), it
           simply returns the content of field MeetingGroup.keepAccessToItemWhenAdviceIsGiven.
           When called with a p_cfg (MeetingConfig), if MeetingGroup.keepAccessToItemWhenAdviceIsGiven
           is '', we will use value defined in MeetingConfig.keepAccessToItemWhenAdviceIsGiven, else
           we will return False if it is '0' and True if it is '1'.'''
        res = self.getField('keepAccessToItemWhenAdviceIsGiven').get(self, **kwargs)
        if cfg:
            if not res:
                res = cfg.getKeepAccessToItemWhenAdviceIsGiven()
            elif res == '0':
                res = False
            else:
                res = True
        return res

    security.declarePublic('getCertifiedSignatures')

    def getCertifiedSignatures(self, computed=False, context=None, from_group_in_charge=False, **kwargs):
        '''Overrides field 'certifiedSignatures' accessor to be able to pass
           the p_computed parameter that will return computed certified signatures,
           so signatures really available right now.  If nothing is defined on the MeetingGroup,
           use certified signatures defined on the corresponding MeetingConfig found using p_context.
           If p_from_group_in_charge is True, we get certifiedSignatures from the first defined
           self.groupsInCharge.'''
        group_signatures = self.getField('certifiedSignatures').get(self, **kwargs)
        if computed:
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(context)
            computedSignatures = cfg.getCertifiedSignatures(computed=True)

            # get certified signatures from first of the defined groupsInCharge
            groups_in_charge = self.getGroupsInCharge()
            if from_group_in_charge and groups_in_charge:
                tool = api.portal.get_tool('portal_plonemeeting')
                group_in_charge = getattr(tool, groups_in_charge[0])
                computedSignatures.update(computeCertifiedSignatures(group_in_charge.getCertifiedSignatures()))

            # if we have certified signatures defined on this MeetingGroup
            # update MeetingConfig signatures regarding what is defined here
            if group_signatures:
                computedSignatures.update(computeCertifiedSignatures(group_signatures))
            # listify signatures, for backward compatibility, we need a list of pair
            # of function/name, like ['function1', 'name1', 'function2', 'name2']
            group_signatures = listifySignatures(computedSignatures)
        return group_signatures


registerType(MeetingGroup, PROJECTNAME)



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from plone import api
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.external.utils import send_json_request
from Products.PloneMeeting.interfaces import IRedirect
from Products.PloneMeeting.widgets.pm_checkbox import PMCheckBoxFieldWidget
from z3c.form import button
from z3c.form import field
from z3c.form import form
from zope import schema
from zope.interface import Interface
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory


@provider(IContextAwareDefaultFactory)
def projects_default(context):
    """ """
    content = send_json_request(
        "delib-links", extra_parameters={"delib_uid": context.UID()})
    return [elt['target']['id'] for elt in content
            if elt['target']['type'] == "project"]


@provider(IContextAwareDefaultFactory)
def tasks_default(context):
    """ """
    content = send_json_request(
        "delib-links", extra_parameters={"delib_uid": context.UID()})
    return [elt['target']['id'] for elt in content
            if elt['target']['type'] == "task"]


class ILinkWithVision(Interface):

    quick_filter = schema.TextLine(
        title=_(u"Quick filter"),
        description=u"",
        required=False,
    )

    projects = schema.List(
        title=_(u"Projects"),
        description=u"",
        required=True,
        defaultFactory=projects_default,
        value_type=schema.Choice(
          vocabulary=u'Products.PloneMeeting.external.vocabularies.visionprojects'),
    )

    tasks = schema.List(
        title=_(u"Tasks"),
        description=u"",
        required=True,
        defaultFactory=tasks_default,
        value_type=schema.Choice(
          vocabulary=u'Products.PloneMeeting.external.vocabularies.visiontasks'),
    )


class LinkWithVisionForm(form.Form):

    fields = field.Fields(ILinkWithVision)
    fields['projects'].widgetFactory = PMCheckBoxFieldWidget
    fields['tasks'].widgetFactory = PMCheckBoxFieldWidget

    ignoreContext = True  # don't use context to get widget data

    label = _(u"Link with iA.Vision")
    description = _(u"Elements displayed are only elements you can access in iA.Vision.")
    _finished_sent = False

    def update(self):
        """ """
        if not self.context.restrictedTraverse('@@load-external-infos').can_link():
            raise Unauthorized
        super(LinkWithVisionForm, self).update()
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')

    def updateWidgets(self):
        super(LinkWithVisionForm, self).updateWidgets()
        self.widgets['quick_filter'].placeholder = _(u'Encode terms to filter')
        # enable filterByName
        self.widgets['quick_filter'].onkeyup = u'filterByName(event)'
        # disable submit on [Enter]
        self.widgets['quick_filter'].onkeydown=u"return (event.keyCode!=13);"

    @button.buttonAndHandler(_('Apply'), name='apply')
    def handle_apply(self, action):
        if not self.context.restrictedTraverse('@@load-external-infos').can_link():
            raise Unauthorized
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        # do link/unlink with iA.Vision
        self._do_link_unlink(data)

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self._finished_sent = True

    def render(self):
        if self._finished_sent:
            IRedirect(self.request).redirect(
                self.context.absolute_url() +
                "?expand-collapsible=external-infos")
            return ""
        return super(LinkWithVisionForm, self).render()

    def _do_link_unlink(self, data):
        """ """
        # get linked items and link new elements and unlink no more selected ones
        linked_content = send_json_request(
            "delib-links", extra_parameters={"delib_uid": self.context.UID()})
        linked_projects = [elt['target']['id'] for elt in linked_content
                           if elt['target']['type'] == "project"]
        linked_tasks = [elt['target']['id'] for elt in linked_content
                        if elt['target']['type'] == "task"]
        # link new selected elements
        for project_id in data['projects']:
            if project_id not in linked_projects:
                project_data = {
                    "delib_uid": self.context.UID(),
                    "target": {"type": "project",
                               "object_id": project_id}}
                res = send_json_request("delib-links", method='POST', data=project_data)
                api.portal.show_message(
                    _('Element "${element}" has been linked.',
                      mapping={'element': res[0]['target']['name']}),
                    request=self.request)
        for task_id in data['tasks']:
            if task_id not in linked_tasks:
                task_data = {
                    "delib_uid": self.context.UID(),
                    "target": {"type": "task",
                               "object_id": task_id}}
                res = send_json_request("delib-links", method='POST', data=task_data)
                api.portal.show_message(
                    _('Element "${element}" has been linked.',
                      mapping={'element': res[0]['target']['name']}),
                    request=self.request)
        # unlink no more selected elements
        for linked_elt in linked_content:
            if (linked_elt['target']['type'] == "project" and
                linked_elt['target']['id'] not in data['projects']) or \
               (linked_elt['target']['type'] == "task" and
                    linked_elt['target']['id'] not in data['tasks']):
                res = send_json_request("delib-links/%s" % linked_elt['id'], method='DELETE')
                api.portal.show_message(
                    _('Element "${element}" has been unlinked.',
                      mapping={'element': linked_elt['target']['name']}),
                    request=self.request,
                    type='warning')
        self._finished_sent = True



# -*- coding: utf-8 -*-

from DateTime import DateTime
from imio.history.utils import add_event_to_history
from persistent.list import PersistentList
from plone import api
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFPlone.utils import base_hasattr
from Products.PloneMeeting.content.advice import IMeetingAdvice
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.PloneMeeting.setuphandlers import _configurePortalRepository
from Products.PloneMeeting.utils import get_dx_data
from Products.PloneMeeting.utils import getAdvicePortalTypeIds
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.event import notify


class Migrate_To_4205(Migrator):

    def _updateConfigCommitteesAndVotesResult(self):
        """MeetingConfig.committees get a new value "enable_groups" and
           init new field MeetingItem.votesResult."""
        logger.info('Updating datagridfield "committees" for every MeetingConfigs...')
        # reinstall workflows to take new role "MeetingCommitteeEditor" into account
        self.runProfileSteps('Products.PloneMeeting', steps=['rolemap', 'workflow'], profile='default')
        for cfg in self.tool.objectValues('MeetingConfig'):
            committees = cfg.getCommittees()
            for committee in committees:
                if "enable_editors" not in committee or \
                   not committee["enable_editors"]:
                    committee["enable_editors"] = "0"
            cfg.setCommittees(committees)
            notify(ObjectEditedEvent(cfg))
        # update new fields committeeTranscript and votesResult on items
        self.initNewHTMLFields(
            query={'meta_type': ('MeetingItem')},
            field_names=('committeeTranscript', 'votesResult'))
        logger.info('Done.')

    def _updateMeetingCommittees(self):
        """Initialize "committee_observations" column for every meetings..."""
        logger.info('Initializing "committee_observations" for every meetings "committees"...')
        brains = self.catalog(object_provides=IMeeting.__identifier__)
        for brain in brains:
            meeting = brain.getObject()
            committees = meeting.committees
            if not committees or "committee_observations" in committees[0]:
                continue
            for committee in committees:
                committee["committee_observations"] = None
            meeting.committees = committees
            meeting._p_changed = True
        logger.info('Done.')

    def _initAdviceGivenHistory(self):
        """Moving from versioning to imio.history advice_given_history."""
        # if using MeetingConfig.enableAdviceInvalidation, clean potential orphan brains
        # enableAdviceInvalidation is disabled during migrations and stored in cfgsAdvicesInvalidation
        if True in self.cfgsAdvicesInvalidation.values():
            self.clean_orphan_brains(
                query={"object_provides": IMeetingAdvice.__identifier__})

        pr = api.portal.get_tool('portal_repository')
        phs = api.portal.get_tool('portal_historiesstorage')
        h_repo = phs._getZVCRepo()
        logger.info('Moving to "advice_given_history"...')
        brains = self.catalog(object_provides=IMeetingAdvice.__identifier__)
        pghandler = ZLogHandler(steps=1000)
        pghandler.init('Moving to "advice_given_history"...',
                       len(brains))
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            advice = brain.getObject()
            if base_hasattr(advice, "advice_given_history"):
                continue
            advice.advice_given_history = PersistentList()
            h_metadata = pr.getHistoryMetadata(advice)
            if h_metadata:
                for version_info in h_metadata._full.values():
                    vc_info = version_info['vc_info']
                    version_advice = h_repo._histories[vc_info.history_id]._versions[
                        vc_info.version_id]._data._object.object
                    if version_advice:
                        # save info in advice_given_history attribute
                        item_data = list(getattr(version_advice, "historized_item_data", []))
                        # get_dx_data does not work because version_advice
                        # is an Acquisition.ImplicitAcquisitionWrapper without acquisition
                        version_advice.aq_parent = advice.aq_parent
                        version_advice.REQUEST = self.request
                        advice_data = get_dx_data(version_advice)
                        version_advice.aq_parent = None
                        version_advice.REQUEST = None
                        meta = version_info["metadata"]["sys_metadata"]
                        # replace some automatic comments
                        comment = meta["comment"]
                        comment = comment.replace("Versioned", "Historized")
                        add_event_to_history(
                            advice,
                            'advice_given_history',
                            action='advice_given_or_modified',
                            actor=api.user.get(meta["principal"]),
                            time=DateTime(meta["timestamp"]),
                            comments=comment,
                            extra_infos={'item_data': item_data,
                                         'advice_data': advice_data})
                    else:
                        logger.info('Could not find version_id {0} for advice version of {1}!'.format(
                            vc_info.version_id, advice.absolute_url_path()))

        # nothing should be versionned anymore
        if self.portal.portal_historiesstorage._shadowStorage is not None:
            self.portal.portal_historiesstorage._shadowStorage._storage.clear()
        self.portal.portal_historiesstorage.zvc_repo._histories.clear()
        # remove every meetingadvice portal_types from portal_repository
        _configurePortalRepository(removed_types=getAdvicePortalTypeIds())
        # MeetingConfig.versionateAdviceIfGivenAndItemModified was renamed to
        # MeetingConfig.historizeAdviceIfGivenAndItemModified
        self.cleanMeetingConfigs(
            field_names=['versionateAdviceIfGivenAndItemModified'],
            renamed={
                'versionateAdviceIfGivenAndItemModified':
                'historizeAdviceIfGivenAndItemModified'})

    def _updateLocalRolesItemBeforeStateValidated(self):
        """Add permissions were not correctly setup on items in state before "validated"."""
        logger.info('Updating "Add permissions" for every items before "validated"...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            review_states = cfg.getItemWFValidationLevels(data='state', only_enabled=True)
            brains = self.catalog(portal_type=cfg.getItemTypeName(), review_state=review_states)
            if brains:
                self.tool.update_all_local_roles(brains=brains)
        logger.info('Done.')

    def _removeCfgUseGroupsAsCategories(self):
        """Field MeetingConfig.useGroupsAsCategories was removed,
           enable 'category' in MeetingConfig.usedItemAttributes when relevant."""
        logger.info('Removing field "MeetingConfig.useGroupsAsCategories" from every MeetingConfigs...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            if not base_hasattr(cfg, 'useGroupsAsCategories'):
                continue
            enable_category = not cfg.useGroupsAsCategories
            if enable_category:
                logger.info("'category' was enabled for MeetingConfig %s" % cfg.getId())
                used_item_attrs = list(cfg.getUsedItemAttributes())
                used_item_attrs.append('category')
                cfg.setUsedItemAttributes(used_item_attrs)
            delattr(cfg, 'useGroupsAsCategories')
        logger.info('Done.')

    def run(self, extra_omitted=[], from_migration_to_4200=False):

        logger.info('Migrating to PloneMeeting 4205...')

        # not necessary if executing the full upgrade to 4200
        if not from_migration_to_4200:
            # will upgrade collective.documentgenerator and collective.messagesviewlet
            self.upgradeAll(omit=['Products.PloneMeeting:default'])
            self._updateConfigCommitteesAndVotesResult()
            self._updateMeetingCommittees()
            self._updateLocalRolesItemBeforeStateValidated()
            self.addNewSearches()

        self._initAdviceGivenHistory()
        self._removeCfgUseGroupsAsCategories()
        logger.info('Migrating to PloneMeeting 4205... Done.')


def migrate(context):
    '''This migration function will:

       1) Update MeetingConfig.committees to add "enable_groups",
          init new MeetingItem fields 'committeeTranscript' and 'votesResult';
       2) Update meetig.committees to add "committee_observations";
       3) Update local roles of items in before review_state "validated";
       4) Move from advice versioning to "advice_given_history";
       5) Remove field "MeetingConfig.useGroupsAsCategories".
    '''
    migrator = Migrate_To_4205(context)
    migrator.run()
    migrator.finish()



from Products.Archetypes.event import EditBegunEvent
from Products.Archetypes.event import EditCancelledEvent
from Products.Five import BrowserView
from zope.event import notify


class NotifyEvent(BrowserView):
    """This is used to notify some events from various places."""

    def notifyEditBegunEvent(self):
        notify(EditBegunEvent(self.context))

    def notifyEditCancelledEvent(self):
        notify(EditCancelledEvent(self.context))



# -*- coding: utf-8 -*-

from collective.iconifiedcategory.browser.actionview import BaseView as BaseActionView
from collective.iconifiedcategory.browser.actionview import ConfidentialChangeView
from collective.iconifiedcategory.browser.actionview import PublishableChangeView
from collective.iconifiedcategory.browser.actionview import SignedChangeView
from collective.iconifiedcategory.browser.actionview import ToPrintChangeView
from collective.iconifiedcategory.browser.tabview import CategorizedTable
from collective.iconifiedcategory.browser.tabview import CategorizedTabView
from collective.iconifiedcategory.browser.views import CategorizedChildInfosView
from collective.iconifiedcategory.browser.views import CategorizedChildView
from collective.iconifiedcategory.utils import _categorized_elements
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.content import get_vocab
from plone import api
from plone.memoize import ram
from Products.PloneMeeting.browser.overrides import BaseActionsPanelView
from Products.PloneMeeting.config import FACETED_ANNEXES_CRITERION_ID
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import get_annexes
from Products.PloneMeeting.utils import get_annexes_config


class AnnexActionsPanelView(BaseActionsPanelView):
    """
      Specific actions displayed on an annex.
    """

    def showHistoryForContext(self):
        """
          History on annexes is only shown to Managers.
        """
        if not super(AnnexActionsPanelView, self).showHistoryForContext():
            return False

        # check isManager on parent (item) so caching is used as context is a key
        # used in the caching invalidation
        parent = self.context.aq_inner.aq_parent
        while parent.__class__.__name__ not in ('MeetingItem', 'Meeting'):
            parent = parent.aq_inner.aq_parent
        return self.tool.isManager(realManagers=True)


class CategorizedAnnexesTable(CategorizedTable):
    """ """

    def initColumns(self):
        """Change name of checkbox column for annex decisions,
           because for batchaction, column name must be different than
           checkbox column of normal annexes.
           """
        super(CategorizedAnnexesTable, self).initColumns()
        if self.portal_type == 'annexDecision':
            column = self.columnByName['select_row']
            column.name = "select_item_annex_decision"


class CategorizedAnnexesView(CategorizedTabView):
    """ """

    def __call__(self):
        """ """
        self._update()
        return super(CategorizedAnnexesView, self).__call__()

    def _update(self):
        """ """
        self.portal_url = api.portal.get().absolute_url()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        # compute show annexes here and display a message to the (Meeting)Managers
        # if not able to add annexes because nothing found in annex vocab nor annexDecision vocab
        annex_vocab, self._showAddAnnex = self.showAddAnnex()
        self.annexDecision_vocab, self._showAddAnnexDecision = self.showAddAnnexDecision()
        if not len(annex_vocab) and not len(self.annexDecision_vocab) and self.tool.isManager(self.cfg):
            api.portal.show_message(
                _('The configuration does not let you add annexes.'),
                request=self.request)

    def _check_can_view(self):
        """ """
        return self.context.__class__.__name__ == "Meeting"

    def _config(self):
        """ """
        return get_annexes_config(self.context, self.portal_type)

    def _show_column(self, action_type):
        """Made to be overrided."""
        annex_attr_config = '{0}_display'.format(action_type)
        check = annex_attr_config in self.cfg.getAnnexRestrictShownAndEditableAttributes()
        return not check or self.tool.isManager(self.cfg)

    def showAddAnnex(self):
        """ """
        portal_types = api.portal.get_tool('portal_types')
        annexTypeInfo = portal_types['annex']
        vocab = get_vocab(self.context, 'collective.iconifiedcategory.categories')
        show = annexTypeInfo in self.context.allowedContentTypes() and len(vocab)
        return vocab, show

    def showAddAnnexDecision(self):
        """ """
        portal_types = api.portal.get_tool('portal_types')
        annexTypeInfo = portal_types['annexDecision']
        self.request.set('force_use_item_decision_annexes_group', True)
        vocab = get_vocab(self.context, 'collective.iconifiedcategory.categories')
        self.request.set('force_use_item_decision_annexes_group', False)
        show = annexTypeInfo in self.context.allowedContentTypes() and len(vocab)
        return vocab, show

    def showAnnexesSection(self):
        """Always show this section, a message is displayed in case configuration
           is not correct, this invite Managers to use annexes."""
        return True

    def showDecisionAnnexesSection(self):
        """Check if context contains decisionAnnexes or if there
           are some decisionAnnex annex types available in the configuration."""
        if self.context.__class__.__name__ == 'MeetingItem' and \
            (get_annexes(self.context, portal_types=['annexDecision']) or
             self._showAddAnnexDecision or len(self.annexDecision_vocab)):
            return True
        return False


def _get_filters(request):
    """ """
    # caching
    res = request.get("cached_annexes_filters", None)
    if res is None:
        res = {}
        # in request.form, faceted criterion is like 'c20[]'
        faceted_filter = request.form.get(FACETED_ANNEXES_CRITERION_ID + '[]', None)
        if faceted_filter is not None:
            if not hasattr(faceted_filter, '__iter__'):
                faceted_filter = [faceted_filter]
            for value in faceted_filter:
                if value.startswith('not_'):
                    res[value.replace('not_', '')] = False
                else:
                    res[value] = True
        request["cached_annexes_filters"] = res
    return res


class PMCategorizedChildView(CategorizedChildView):
    """Add caching."""

    def __call___cachekey(method, self, portal_type=None, show_nothing=False, check_can_view=False):
        '''cachekey method for self.__call__.'''
        categorized_elements = _categorized_elements(self.context)
        if not categorized_elements:
            return show_nothing
        # value of the annexes faceted filter
        filters = self._filters
        if filters:
            raise ram.DontCache

        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        # URL to the annex_type can change if server URL changed
        server_url = self.request.get('SERVER_URL', None)
        # use "last_updated" to know when categorized_elements was last updated...
        last_updated = max(
            [info['last_updated'] for info in categorized_elements.values()])
        # this is necessary in case an annex is removed
        len_annexes = len(categorized_elements)
        cfg_modified = cfg.modified()
        # check confidential annexes if not MeetingManager
        isManager = tool.isManager(cfg)
        # manage confidential annexes for non managers
        may_view_confidential_annexes = True
        if not isManager:
            confidential_annexes = [
                elt for elt in self.context.categorized_elements.values()
                if elt["confidential"]]
            may_view_confidential_annexes = not confidential_annexes or \
                set(get_plone_groups_for_user()).intersection(
                    confidential_annexes[0]["visible_for_groups"])
        return (repr(self.context),
                last_updated,
                len_annexes,
                cfg_modified,
                server_url,
                may_view_confidential_annexes,
                portal_type,
                show_nothing,
                check_can_view)

    @property
    def _filters(self):
        """ """
        return _get_filters(self.request)

    @ram.cache(__call___cachekey)
    def __call__(self, portal_type=None, show_nothing=False, check_can_view=False):
        """Override to change show_nothing=False instead
           show_nothing=True and to add caching."""
        return super(PMCategorizedChildView, self).__call__(
            portal_type, show_nothing, check_can_view)


class PMCategorizedChildInfosView(CategorizedChildInfosView):
    """ """

    def __init__(self, context, request):
        """ """
        super(PMCategorizedChildInfosView, self).__init__(context, request)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    @property
    def _filters(self):
        """ """
        return _get_filters(self.request)

    def show_preview(self, element):
        """Show preview if the auto_convert in collective.documentviewer is enabled."""
        return super(PMCategorizedChildInfosView, self).show_preview(element) or \
            self.tool.auto_convert_annexes()

    def _show_protected_download(self, element):
        """When "show_preview" is "2", trigger advanced check.
           Are allowed to download:
           - proposingGroup members;
           - (Meeting)Managers."""
        return self.tool.isManager(self.cfg) or \
            self.context.getProposingGroup() in self.tool.get_orgs_for_user()

    def show_nothing(self):
        """Do not display the 'Nothing' label."""
        return False

    def categorized_elements_more_infos_url(self):
        """ """
        return "{0}/@@categorized-annexes".format(self.context.absolute_url())

    def _show_detail(self, detail_type):
        """ """
        annex_attr_config = '{0}_display'.format(detail_type)
        check = annex_attr_config in self.cfg.getAnnexRestrictShownAndEditableAttributes()
        return not check or self.tool.isManager(self.cfg)


class PMBaseActionView(BaseActionView):
    """ """

    def _get_config_attributes(self):
        """ """
        category_group_attr_name = getattr(self, 'category_group_attr_name', None)
        # get current type of action
        # turn to_be_printed_activated to to_be_printed_edit/to_be_printed_display
        # if an element is not displayed only to MeetingManagers, we consider it is ony editable as well
        config_attr_display = category_group_attr_name.replace('_activated', '_display')
        config_attr_edit = category_group_attr_name.replace('_activated', '_edit')
        return config_attr_display, config_attr_edit

    def _may_set_values(self, values, ):
        res = super(PMBaseActionView, self)._may_set_values(values)
        if res:
            # get current type of action
            # turn to_be_printed_activated to to_be_printed_edit/to_be_printed_display
            # if an element is not displayed only to MeetingManagers, we consider it is ony editable as well
            config_attr_display, config_attr_edit = self._get_config_attributes()
            # restricted to MeetingManagers?
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self.context)
            annex_attr_config = cfg.getAnnexRestrictShownAndEditableAttributes()
            if config_attr_display in annex_attr_config or config_attr_edit in annex_attr_config:
                res = tool.isManager(cfg)
        return res


class PMConfidentialChangeView(ConfidentialChangeView, PMBaseActionView):
    """Override to use new base klass."""


class PMPublishableChangeView(PublishableChangeView, PMBaseActionView):
    """Override to use new base klass."""


class PMSignedChangeView(SignedChangeView, PMBaseActionView):
    """Override to use new base klass."""


class PMToPrintChangeView(ToPrintChangeView, PMBaseActionView):
    """Override to use new base klass."""
